package dev.fyloz.colorrecipesexplorer.databasemanager

import mu.KotlinLogging
import java.io.File
import kotlin.system.exitProcess

private const val ACTION_ARG_INDEX = 0
private const val PROPERTIES_PATH_ARG_INDEX = 1

private const val ACTION_VERSION_CHECK = "versionCheck"
private const val ACTION_UPDATE = "update"

private var propertiesPath: String? = null
private var database: CreDatabase? = null
private val logger = KotlinLogging.logger { }

// Lazy because propertiesPath is needed to load the properties file
private val properties by lazy { loadProperties(propertiesPath) }

internal fun main(args: Array<String>) {
    logger.info("Starting Color Recipes Explorer database updater...")

    if (args.isEmpty() || args.size < 2) {
        logger.error("No enough parameters were specified. ")
        exitLogUsage()
    }

    val action = args[ACTION_ARG_INDEX]
    propertiesPath = args[PROPERTIES_PATH_ARG_INDEX]

    try {
        database = CreDatabase(DatabaseContext(properties, logger))

        // Run the specified action
        when (action) {
            ACTION_VERSION_CHECK -> executeVersionCheck(database!!)
            ACTION_UPDATE -> executeUpdate(database!!)
            else -> {
                logger.error("Invalid action '$action'")
                exitLogUsage()
            }
        }
    } catch (ex: CreDatabaseException) {
        if (ex is CreDatabaseException.Connection) {
            logger.error(ex.message, ex)
        } else {
            logger.error(ex.message)
        }
        exitError()
    }

    exit()
}

private fun executeVersionCheck(database: CreDatabase) {
    logger.info("Performing a version check...")
    logger.info("Target version: ${properties.targetVersion}")

    val databaseVersion = database.version
    logger.info("Database version: $databaseVersion")

    when (databaseVersion.compareTo(properties.targetVersion)) {
        1 -> logger.info("Database version is higher than target version!")
        0 -> logger.info("Database version matches the target version!")
        -1 -> logger.info("Database version is lower than target version!")
    }
}

private fun executeUpdate(database: CreDatabase) {
    logger.info("Performing an update...")

    database.update()
}

/** Load properties from the given [path]. */
private fun loadProperties(path: String?): DatabaseUpdaterProperties {
    // Stop the utility if propertiesPath is null
    if (path == null) {
        logger.error("No path to a properties file was specified")
        exitError()
    }

    val file = File(path)
    logger.debug("Loading properties from ${file.absolutePath}")
    try {
        val properties = databaseUpdaterProperties(file)
        logger.debug("Loaded properties: $properties")
        return properties
    } catch (exception: InvalidDatabaseUpdaterPropertiesException) {
        logger.error(exception.message)
        exitError()
    }
}

/** Exits the utility. */
private fun exit(): Nothing {
    database?.close()
    logger.info("Exiting Color Recipes Explorer database updater without error")
    exitProcess(0)
}

/** Exits the utility with an error code and a warning. */
private fun exitError(): Nothing {
    database?.close()
    logger.warn("Exiting Color Recipes Explorer database updater with an error")
    exitProcess(1)
}

/** Exits the utility and logs how to use it. */
private fun exitLogUsage(): Nothing {
    logger.error("Usage: java -jar DatabaseUpdater.jar <action> <path to properties>")
    logger.error("Actions: ")
    logger.error("\t- versionCheck: Compare the database's version to the target version specified in the properties")
    logger.error("\t- update: Update the database to the target version specified in the properties")
    exitError()
}
